import { makeColumnTypes, SqlMethod, QueryHooks, defaultSchemaConfig, getColumnTypes, parseTableData, _queryHookAfterCreate, _queryHookAfterUpdate, getQueryAs, setQueryObjectValue, pushQueryOn, VirtualColumn, pushQueryValue, _queryCreateMany, isQueryReturnsAll, _queryHookBeforeUpdate, _queryFindBy, _queryCreate, _queryRows, _queryUpdate, _queryDelete, _queryDefaults, _queryUpdateOrThrow, _queryWhere, toSQLCacheKey, _queryJoinOn, OrchidOrmInternalError, _queryCreateFrom, NotFoundError, _queryFindByOptional, _querySelect, _queryWhereExists, _queryTake, _queryTakeOptional, _queryAll, Adapter, _initQueryBuilder, Db, addComputedColumns, getClonedQueryData } from 'pqb';
export * from 'pqb';
import { getStackTrace, applyMixins, getCallerFilePath, snakeCaseKey, toSnakeCase, emptyObject, emptyArray, objectHasValues, toArray } from 'orchid-core';
export * from 'orchid-core';
import { AsyncLocalStorage } from 'node:async_hooks';

function createBaseTable({
  schemaConfig = defaultSchemaConfig,
  columnTypes: columnTypesArg,
  snakeCase,
  filePath: filePathArg,
  nowSQL,
  exportAs = "BaseTable",
  language
} = {}) {
  var _a;
  const columnTypes = typeof columnTypesArg === "function" ? columnTypesArg(makeColumnTypes(schemaConfig)) : columnTypesArg || makeColumnTypes(schemaConfig);
  const filePathOrStack = filePathArg || getStackTrace();
  let filePath;
  const defaultColumns = {
    shape: emptyObject,
    data: emptyArray
  };
  const base = (_a = class {
    constructor() {
      this.columns = defaultColumns;
      this.snakeCase = snakeCase;
      this.types = columnTypes;
      this.q = {};
      this.language = language;
    }
    static inputSchema() {
      this.instance();
      return this._inputSchema === void 0 ? this._inputSchema = schemaConfig.inputSchema.call(this) : this._inputSchema;
    }
    static outputSchema() {
      this.instance();
      return this._outputSchema === void 0 ? this._outputSchema = schemaConfig.outputSchema.call(this) : this._outputSchema;
    }
    static querySchema() {
      this.instance();
      return this._querySchema === void 0 ? this._querySchema = schemaConfig.querySchema.call(this) : this._querySchema;
    }
    static updateSchema() {
      this.instance();
      return this._updateSchema === void 0 ? this._updateSchema = schemaConfig.updateSchema.call(this) : this._updateSchema;
    }
    static pkeySchema() {
      this.instance();
      return this._pkeySchema === void 0 ? this._pkeySchema = schemaConfig.pkeySchema.call(this) : this._pkeySchema;
    }
    static getFilePath() {
      if (filePath)
        return filePath;
      if (typeof filePathOrStack === "string") {
        return filePath = filePathOrStack;
      }
      filePath = getCallerFilePath(filePathOrStack);
      if (filePath)
        return filePath;
      throw new Error(
        `Failed to determine file path of a base table. Please set the \`filePath\` option of \`createBaseTable\` manually.`
      );
    }
    static instance() {
      var _a2;
      return (_a2 = this._instance) != null ? _a2 : this._instance = new this();
    }
    clone() {
      return this;
    }
    getFilePath() {
      if (this.filePath)
        return this.filePath;
      if (typeof filePathOrStack === "string")
        return this.filePath = filePathOrStack;
      const filePath2 = getCallerFilePath(filePathOrStack);
      if (filePath2)
        return this.filePath = filePath2;
      throw new Error(
        `Failed to determine file path for table ${this.constructor.name}. Please set \`filePath\` property manually`
      );
    }
    setColumns(fn, dataFn) {
      columnTypes[snakeCaseKey] = this.snakeCase;
      const shape = getColumnTypes(columnTypes, fn, nowSQL, this.language);
      const tableData = this.constructor.prototype.tableData = parseTableData(dataFn);
      if (this.snakeCase) {
        for (const key in shape) {
          const column = shape[key];
          if (column.data.name)
            continue;
          const snakeName = toSnakeCase(key);
          if (snakeName !== key) {
            column.data.name = snakeName;
          }
        }
      }
      return this.constructor.prototype.columns = {
        shape,
        data: tableData
      };
    }
    setComputed(computed) {
      return computed;
    }
    setScopes(scopes) {
      return scopes;
    }
    belongsTo(fn, options) {
      return {
        type: "belongsTo",
        fn,
        options
      };
    }
    hasOne(fn, options) {
      return {
        type: "hasOne",
        fn,
        options
      };
    }
    hasMany(fn, options) {
      return {
        type: "hasMany",
        fn,
        options
      };
    }
    hasAndBelongsToMany(fn, options) {
      return {
        type: "hasAndBelongsToMany",
        fn,
        options
      };
    }
  }, _a.nowSQL = nowSQL, _a.exportAs = exportAs, _a.columnTypes = columnTypes, _a.sql = SqlMethod.prototype.sql, _a);
  applyMixins(base, [QueryHooks]);
  base.prototype.types = columnTypes;
  return base;
}

const getThroughRelation = (table, through) => {
  var _a;
  return (_a = table.relations[through]) == null ? void 0 : _a.relationConfig;
};
const getSourceRelation = (throughRelation, source) => {
  var _a;
  return (_a = throughRelation.query.relations[source]) == null ? void 0 : _a.relationConfig;
};
const hasRelationHandleCreate = (q, ctx, item, rowIndex, key, primaryKeys, nestedInsert) => {
  const value = item[key];
  if ((!value.create || Array.isArray(value.create) && value.create.length === 0) && (!value.connect || Array.isArray(value.connect) && value.connect.length === 0) && (!value.connectOrCreate || Array.isArray(value.connectOrCreate) && value.connectOrCreate.length === 0))
    return;
  const store = ctx;
  if (!store.hasRelation)
    store.hasRelation = {};
  const values = [rowIndex, value];
  if (store.hasRelation[key]) {
    store.hasRelation[key].push(values);
    return;
  }
  q.q.wrapInTransaction = true;
  const relationData = [values];
  store.hasRelation[key] = relationData;
  _queryHookAfterCreate(
    q,
    primaryKeys,
    (rows, q2) => nestedInsert(
      q2,
      relationData.map(([rowIndex2, data]) => [
        rows[rowIndex2],
        data
      ])
    )
  );
};
const hasRelationHandleUpdate = (q, set, key, primaryKeys, nestedUpdate) => {
  const value = set[key];
  if (!value.set && !("upsert" in value) && (!value.disconnect || Array.isArray(value.disconnect) && value.disconnect.length === 0) && (!value.delete || Array.isArray(value.delete) && value.delete.length === 0) && (!value.update || Array.isArray(value.update.where) && value.update.where.length === 0) && (!value.create || Array.isArray(value.create) && value.create.length === 0))
    return;
  selectIfNotSelected(q, primaryKeys);
  q.q.wrapInTransaction = true;
  _queryHookAfterUpdate(q, primaryKeys, (rows, q2) => {
    return nestedUpdate(
      q2,
      rows,
      value
    );
  });
};
const selectIfNotSelected = (q, columns) => {
  const select = q.q.select || [];
  if (!select.includes("*")) {
    for (const column of columns) {
      if (!select.includes(column)) {
        select.push(column);
      }
    }
    q.q.select = select;
  }
};
const relationWhere = (len, keys, valueKeys) => (params) => {
  const obj = {};
  for (let i = 0; i < len; i++) {
    obj[keys[i]] = params[valueKeys[i]];
  }
  return obj;
};
function joinHasThrough(q, baseQuery, joiningQuery, throughRelation, sourceRelation) {
  return q.whereExists(
    throughRelation.joinQuery(throughRelation.query, baseQuery),
    () => {
      const as = getQueryAs(joiningQuery);
      return sourceRelation.joinQuery(
        sourceRelation.query.as(as),
        throughRelation.query
      );
    }
  );
}
function joinHasRelation(baseQuery, joiningQuery, primaryKeys, foreignKeys, len) {
  const q = joiningQuery.clone();
  setQueryObjectValue(
    q,
    "joinedShapes",
    baseQuery.q.as || baseQuery.table,
    baseQuery.q.shape
  );
  for (let i = 0; i < len; i++) {
    pushQueryOn(q, baseQuery, joiningQuery, foreignKeys[i], primaryKeys[i]);
  }
  return q;
}
const joinQueryChainingHOF = (reverseJoin, joinQuery) => (joiningQuery, baseQuery) => {
  const chain = joiningQuery.q.relChain;
  if (!chain || chain.length === 1) {
    return joinQuery(joiningQuery, baseQuery);
  }
  const last = chain[chain.length - 1];
  const query = "relationConfig" in last ? last.relationConfig.joinQuery(last, baseQuery) : last;
  return joiningQuery.where({
    EXISTS: { q: reverseJoin(query, joiningQuery) }
  });
};

class BelongsToVirtualColumn extends VirtualColumn {
  constructor(schema, key, state) {
    super(schema);
    this.key = key;
    this.state = state;
    this.nestedInsert = nestedInsert$3(this.state);
    this.nestedUpdate = nestedUpdate$3(this.state);
  }
  create(q, ctx, item, rowIndex) {
    const {
      key,
      state: { primaryKeys, foreignKeys }
    } = this;
    const columnIndexes = foreignKeys.map((key2) => {
      let index = ctx.columns.get(key2);
      if (index === void 0) {
        ctx.columns.set(key2, index = ctx.columns.size);
      }
      return index;
    });
    const store = ctx;
    if (!store.belongsTo)
      store.belongsTo = {};
    const values = [rowIndex, columnIndexes, item[key]];
    if (store.belongsTo[key]) {
      store.belongsTo[key].push(values);
      return;
    }
    const relationData = [values];
    store.belongsTo[key] = relationData;
    q.q.wrapInTransaction = true;
    pushQueryValue(q, "beforeCreate", async (q2) => {
      const inserted = await this.nestedInsert(
        q2,
        relationData.map(([, , data]) => data)
      );
      const { values: values2 } = q2.q;
      for (let i = 0, len = relationData.length; i < len; i++) {
        const [rowIndex2, columnIndexes2] = relationData[i];
        const row = values2[rowIndex2];
        for (let c = 0, len2 = columnIndexes2.length; c < len2; c++) {
          row[columnIndexes2[c]] = inserted[i][primaryKeys[c]];
        }
      }
    });
  }
  update(q, ctx, set) {
    q.q.wrapInTransaction = true;
    const data = set[this.key];
    this.nestedUpdate(q, set, data, ctx);
  }
}
const makeBelongsToMethod = (relation, relationName, query) => {
  const primaryKeys = "columns" in relation.options ? relation.options.references : [relation.options.primaryKey];
  const foreignKeys = "columns" in relation.options ? relation.options.columns : [relation.options.foreignKey];
  const len = primaryKeys.length;
  const state = { query, primaryKeys, foreignKeys, len };
  const makeWhere = relationWhere(len, primaryKeys, foreignKeys);
  const join = (baseQuery, joiningQuery, primaryKeys2, foreignKeys2) => {
    const q = joiningQuery.clone();
    setQueryObjectValue(
      q,
      "joinedShapes",
      baseQuery.q.as || baseQuery.table,
      baseQuery.q.shape
    );
    for (let i = 0; i < len; i++) {
      pushQueryOn(q, baseQuery, joiningQuery, primaryKeys2[i], foreignKeys2[i]);
    }
    return q;
  };
  const reverseJoin = (baseQuery, joiningQuery) => {
    return join(joiningQuery, baseQuery, foreignKeys, primaryKeys);
  };
  return {
    returns: "one",
    method(params) {
      return query.where(makeWhere(params));
    },
    virtualColumn: new BelongsToVirtualColumn(
      defaultSchemaConfig,
      relationName,
      state
    ),
    joinQuery: joinQueryChainingHOF(
      reverseJoin,
      (joiningQuery, baseQuery) => join(baseQuery, joiningQuery, primaryKeys, foreignKeys)
    ),
    reverseJoin
  };
};
const nestedInsert$3 = ({ query, primaryKeys }) => {
  return async (_, data) => {
    const t = query.clone();
    const items = [];
    for (const item of data) {
      if (item.connectOrCreate) {
        items.push(item);
      }
    }
    let connectOrCreated;
    if (items.length) {
      for (let i = 0, len = items.length; i < len; i++) {
        items[i] = t.findByOptional(
          items[i].connectOrCreate.where
        );
      }
      connectOrCreated = await Promise.all(items);
    } else {
      connectOrCreated = emptyArray;
    }
    let connectOrCreatedI = 0;
    items.length = 0;
    for (const item of data) {
      if (item.connectOrCreate) {
        if (!connectOrCreated[connectOrCreatedI++])
          items.push(item);
      } else if (item.create) {
        items.push(item);
      }
    }
    let created;
    if (items.length) {
      for (let i = 0, len = items.length; i < len; i++) {
        items[i] = "create" in items[i] ? items[i].create : items[i].connectOrCreate.create;
      }
      created = await _queryCreateMany(
        t.select(...primaryKeys),
        items
      );
    } else {
      created = emptyArray;
    }
    items.length = 0;
    for (const item of data) {
      if (item.connect) {
        items.push(item);
      }
    }
    let connected;
    if (items.length) {
      for (let i = 0, len = items.length; i < len; i++) {
        items[i] = t.findBy(items[i].connect);
      }
      connected = await Promise.all(items);
    } else {
      connected = emptyArray;
    }
    let createdI = 0;
    let connectedI = 0;
    connectOrCreatedI = 0;
    return data.map((item) => {
      return item.connectOrCreate ? connectOrCreated[connectOrCreatedI++] || created[createdI++] : item.connect ? connected[connectedI++] : created[createdI++];
    });
  };
};
const nestedUpdate$3 = ({ query, primaryKeys, foreignKeys, len }) => {
  return (q, update, params, state) => {
    var _a;
    if (params.upsert && isQueryReturnsAll(q)) {
      throw new Error("`upsert` option is not allowed in a batch update");
    }
    let idsForDelete;
    _queryHookBeforeUpdate(q, async (q2) => {
      if (params.disconnect) {
        for (const key of foreignKeys) {
          update[key] = null;
        }
      } else if (params.set) {
        let loadPrimaryKeys;
        let loadForeignKeys;
        for (let i = 0; i < len; i++) {
          const primaryKey = primaryKeys[i];
          if (primaryKey in params.set) {
            update[foreignKeys[i]] = params.set[primaryKey];
          } else {
            (loadPrimaryKeys != null ? loadPrimaryKeys : loadPrimaryKeys = []).push(primaryKey);
            (loadForeignKeys != null ? loadForeignKeys : loadForeignKeys = []).push(foreignKeys[i]);
          }
        }
        if (loadPrimaryKeys) {
          const record = await _queryFindBy(query.select(...loadPrimaryKeys), [
            params.set
          ]);
          for (let i = 0, len2 = loadPrimaryKeys.length; i < len2; i++) {
            update[loadForeignKeys[i]] = record[loadPrimaryKeys[i]];
          }
        }
      } else if (params.create) {
        const q3 = query.clone();
        q3.q.select = primaryKeys;
        const record = await _queryCreate(q3, params.create);
        for (let i = 0; i < len; i++) {
          update[foreignKeys[i]] = record[primaryKeys[i]];
        }
      } else if (params.delete) {
        const selectQuery = q2.clone();
        selectQuery.q.type = void 0;
        selectQuery.q.distinct = emptyArray;
        idsForDelete = await _queryRows(selectQuery);
        for (const foreignKey of foreignKeys) {
          update[foreignKey] = null;
        }
      }
    });
    const { upsert } = params;
    if (upsert || params.update || params.delete) {
      selectIfNotSelected(q, foreignKeys);
    }
    if (upsert) {
      ((_a = state.queries) != null ? _a : state.queries = []).push(async (queryResult) => {
        const row = queryResult.rows[0];
        let obj = {};
        for (let i = 0; i < len; i++) {
          const id = row[foreignKeys[i]];
          if (id === null) {
            obj = void 0;
            break;
          }
          obj[primaryKeys[i]] = id;
        }
        if (obj) {
          await _queryUpdate(
            query.findBy(obj),
            upsert.update
          );
        } else {
          const data = typeof upsert.create === "function" ? upsert.create() : upsert.create;
          const result = await _queryCreate(query.select(...primaryKeys), data);
          const collectData = {};
          state.collect = {
            keys: primaryKeys,
            data: collectData
          };
          for (let i = 0; i < len; i++) {
            collectData[foreignKeys[i]] = result[primaryKeys[i]];
          }
        }
      });
    } else if (params.delete || params.update) {
      _queryHookAfterUpdate(q, [], async (data) => {
        let ids;
        if (params.delete) {
          ids = idsForDelete;
        } else {
          ids = [];
          for (const item of data) {
            let row;
            for (const foreignKey of foreignKeys) {
              const id = item[foreignKey];
              if (id === null) {
                row = void 0;
                break;
              } else {
                (row != null ? row : row = []).push(id);
              }
            }
            if (row)
              ids.push(row);
          }
        }
        if (!(ids == null ? void 0 : ids.length))
          return;
        const t = query.whereIn(
          primaryKeys,
          ids
        );
        if (params.delete) {
          await _queryDelete(t);
        } else {
          await _queryUpdate(t, params.update);
        }
      });
    }
  };
};

var __defProp$5 = Object.defineProperty;
var __getOwnPropSymbols$5 = Object.getOwnPropertySymbols;
var __hasOwnProp$5 = Object.prototype.hasOwnProperty;
var __propIsEnum$5 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$5 = (obj, key, value) => key in obj ? __defProp$5(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$5 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$5.call(b, prop))
      __defNormalProp$5(a, prop, b[prop]);
  if (__getOwnPropSymbols$5)
    for (var prop of __getOwnPropSymbols$5(b)) {
      if (__propIsEnum$5.call(b, prop))
        __defNormalProp$5(a, prop, b[prop]);
    }
  return a;
};
class HasOneVirtualColumn extends VirtualColumn {
  constructor(schema, key, state) {
    super(schema);
    this.key = key;
    this.state = state;
    this.nestedInsert = nestedInsert$2(state);
    this.nestedUpdate = nestedUpdate$2(state);
  }
  create(q, ctx, item, rowIndex) {
    hasRelationHandleCreate(
      q,
      ctx,
      item,
      rowIndex,
      this.key,
      this.state.primaryKeys,
      this.nestedInsert
    );
  }
  update(q, _, set) {
    const params = set[this.key];
    if ((params.set || params.create || params.upsert) && isQueryReturnsAll(q)) {
      const key = params.set ? "set" : params.create ? "create" : "upsert";
      throw new Error(`\`${key}\` option is not allowed in a batch update`);
    }
    hasRelationHandleUpdate(
      q,
      set,
      this.key,
      this.state.primaryKeys,
      this.nestedUpdate
    );
  }
}
const makeHasOneMethod = (table, relation, relationName, query) => {
  if ("through" in relation.options) {
    const { through, source } = relation.options;
    const throughRelation = getThroughRelation(table, through);
    const sourceRelation = getSourceRelation(throughRelation, source);
    const sourceQuery = sourceRelation.joinQuery(sourceRelation.query, throughRelation.query).as(relationName);
    const whereExistsCallback = () => sourceQuery;
    const reverseJoin2 = (baseQuery, joiningQuery) => {
      return joinHasThrough(
        baseQuery,
        baseQuery,
        joiningQuery,
        throughRelation,
        sourceRelation
      );
    };
    return {
      returns: "one",
      method: (params) => {
        const throughQuery = table[through](params);
        return query.whereExists(
          throughQuery,
          whereExistsCallback
        );
      },
      joinQuery: joinQueryChainingHOF(
        reverseJoin2,
        (joiningQuery, baseQuery) => joinHasThrough(
          joiningQuery,
          baseQuery,
          joiningQuery,
          throughRelation,
          sourceRelation
        )
      ),
      reverseJoin: reverseJoin2
    };
  }
  const primaryKeys = "columns" in relation.options ? relation.options.columns : [relation.options.primaryKey];
  const foreignKeys = "columns" in relation.options ? relation.options.references : [relation.options.foreignKey];
  const state = { query, primaryKeys, foreignKeys };
  const len = primaryKeys.length;
  const reversedOn = {};
  for (let i = 0; i < len; i++) {
    reversedOn[foreignKeys[i]] = primaryKeys[i];
  }
  const fromQuerySelect = [{ selectAs: reversedOn }];
  const reverseJoin = (baseQuery, joiningQuery) => {
    return joinHasRelation(
      joiningQuery,
      baseQuery,
      foreignKeys,
      primaryKeys,
      len
    );
  };
  return {
    returns: "one",
    method: (params) => {
      const values = {};
      for (let i = 0; i < len; i++) {
        values[foreignKeys[i]] = params[primaryKeys[i]];
      }
      return _queryDefaults(query.where(values), values);
    },
    virtualColumn: new HasOneVirtualColumn(
      defaultSchemaConfig,
      relationName,
      state
    ),
    joinQuery: joinQueryChainingHOF(
      reverseJoin,
      (joiningQuery, baseQuery) => joinHasRelation(baseQuery, joiningQuery, primaryKeys, foreignKeys, len)
    ),
    reverseJoin,
    modifyRelatedQuery(relationQuery) {
      return (query2) => {
        const baseQuery = query2.clone();
        baseQuery.q.select = fromQuerySelect;
        const q = relationQuery.q;
        q.kind = "from";
        q.values = { from: baseQuery };
      };
    }
  };
};
const nestedInsert$2 = ({ query, primaryKeys, foreignKeys }) => {
  return async (_, data) => {
    const t = query.clone();
    const items = [];
    for (const item of data) {
      if (item[1].connect || item[1].connectOrCreate) {
        items.push(item);
      }
    }
    let connected;
    if (items.length) {
      for (let i = 0, len = items.length; i < len; i++) {
        const [selfData, item] = items[i];
        const data2 = {};
        for (let i2 = 0; i2 < len; i2++) {
          data2[foreignKeys[i2]] = selfData[primaryKeys[i2]];
        }
        items[i] = "connect" in item ? _queryUpdateOrThrow(
          t.where(item.connect),
          data2
        ) : _queryUpdate(
          t.where(
            item.connectOrCreate.where
          ),
          data2
        );
      }
      connected = await Promise.all(items);
    } else {
      connected = [];
    }
    let connectedI = 0;
    items.length = 0;
    for (const item of data) {
      if (item[1].connectOrCreate) {
        if (!connected[connectedI++]) {
          items.push(item);
        }
      } else if (item[1].create) {
        items.push(item);
      }
    }
    if (items.length) {
      for (let i = 0, len = items.length; i < len; i++) {
        const [selfData, item] = items[i];
        const data2 = __spreadValues$5({}, "create" in item ? item.create : item.connectOrCreate.create);
        for (let i2 = 0; i2 < primaryKeys.length; i2++) {
          data2[foreignKeys[i2]] = selfData[primaryKeys[i2]];
        }
        items[i] = data2;
      }
      await t.insertMany(items);
    }
  };
};
const nestedUpdate$2 = ({ query, primaryKeys, foreignKeys }) => {
  const len = primaryKeys.length;
  const setNulls = {};
  for (const foreignKey of foreignKeys) {
    setNulls[foreignKey] = null;
  }
  return async (_, data, params) => {
    const t = query.clone();
    const ids = data.map(
      (item) => primaryKeys.map((primaryKey) => item[primaryKey])
    );
    const currentRelationsQuery = t.whereIn(
      foreignKeys,
      ids
    );
    if (params.create || params.disconnect || params.set) {
      await _queryUpdate(currentRelationsQuery, setNulls);
      const record = data[0];
      if (params.create) {
        const obj = __spreadValues$5({}, params.create);
        for (let i = 0; i < len; i++) {
          obj[foreignKeys[i]] = record[primaryKeys[i]];
        }
        await t.insert(obj);
      }
      if (params.set) {
        const obj = {};
        for (let i = 0; i < len; i++) {
          obj[foreignKeys[i]] = record[primaryKeys[i]];
        }
        await _queryUpdate(
          _queryWhere(t, [params.set]),
          obj
        );
      }
    } else if (params.update) {
      await _queryUpdate(currentRelationsQuery, params.update);
    } else if (params.delete) {
      await _queryDelete(currentRelationsQuery);
    } else if (params.upsert) {
      const { update, create } = params.upsert;
      currentRelationsQuery.q.select = foreignKeys;
      const updatedIds = await _queryUpdate(
        _queryRows(currentRelationsQuery),
        update
      );
      if (updatedIds.length < ids.length) {
        const data2 = typeof create === "function" ? create() : create;
        await t.createMany(
          ids.reduce((rows, ids2) => {
            if (!updatedIds.some(
              (updated) => updated.every((value, i) => value === ids2[i])
            )) {
              const obj = __spreadValues$5({}, data2);
              for (let i = 0; i < len; i++) {
                obj[foreignKeys[i]] = ids2[i];
              }
              rows.push(obj);
            }
            return rows;
          }, [])
        );
      }
    }
  };
};

var __defProp$4 = Object.defineProperty;
var __getOwnPropSymbols$4 = Object.getOwnPropertySymbols;
var __hasOwnProp$4 = Object.prototype.hasOwnProperty;
var __propIsEnum$4 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$4 = (obj, key, value) => key in obj ? __defProp$4(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$4 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$4.call(b, prop))
      __defNormalProp$4(a, prop, b[prop]);
  if (__getOwnPropSymbols$4)
    for (var prop of __getOwnPropSymbols$4(b)) {
      if (__propIsEnum$4.call(b, prop))
        __defNormalProp$4(a, prop, b[prop]);
    }
  return a;
};
class HasManyVirtualColumn extends VirtualColumn {
  constructor(schema, key, state) {
    super(schema);
    this.key = key;
    this.state = state;
    this.nestedInsert = nestedInsert$1(state);
    this.nestedUpdate = nestedUpdate$1(state);
  }
  create(q, ctx, item, rowIndex) {
    hasRelationHandleCreate(
      q,
      ctx,
      item,
      rowIndex,
      this.key,
      this.state.primaryKeys,
      this.nestedInsert
    );
  }
  update(q, _, set) {
    const params = set[this.key];
    if ((params.set || params.create) && isQueryReturnsAll(q)) {
      const key = params.set ? "set" : "create";
      throw new Error(`\`${key}\` option is not allowed in a batch update`);
    }
    hasRelationHandleUpdate(
      q,
      set,
      this.key,
      this.state.primaryKeys,
      this.nestedUpdate
    );
  }
}
const makeHasManyMethod = (table, relation, relationName, query) => {
  if ("through" in relation.options) {
    const { through, source } = relation.options;
    const throughRelation = getThroughRelation(table, through);
    const sourceRelation = getSourceRelation(throughRelation, source);
    const sourceRelationQuery = sourceRelation.query.as(relationName);
    const sourceQuery = sourceRelation.joinQuery(
      sourceRelationQuery,
      throughRelation.query
    );
    const whereExistsCallback = () => sourceQuery;
    const reverseJoin2 = (baseQuery, joiningQuery) => {
      return joinHasThrough(
        baseQuery,
        baseQuery,
        joiningQuery,
        throughRelation,
        sourceRelation
      );
    };
    return {
      returns: "many",
      method: (params) => {
        const throughQuery = table[through](params);
        return query.whereExists(
          throughQuery,
          whereExistsCallback
        );
      },
      joinQuery: joinQueryChainingHOF(
        reverseJoin2,
        (joiningQuery, baseQuery) => joinHasThrough(
          joiningQuery,
          baseQuery,
          joiningQuery,
          throughRelation,
          sourceRelation
        )
      ),
      reverseJoin: reverseJoin2
    };
  }
  const primaryKeys = "columns" in relation.options ? relation.options.columns : [relation.options.primaryKey];
  const foreignKeys = "columns" in relation.options ? relation.options.references : [relation.options.foreignKey];
  const state = { query, primaryKeys, foreignKeys };
  const len = primaryKeys.length;
  const reversedOn = {};
  for (let i = 0; i < len; i++) {
    reversedOn[foreignKeys[i]] = primaryKeys[i];
  }
  const fromQuerySelect = [{ selectAs: reversedOn }];
  const reverseJoin = (baseQuery, joiningQuery) => {
    return joinHasRelation(
      joiningQuery,
      baseQuery,
      foreignKeys,
      primaryKeys,
      len
    );
  };
  return {
    returns: "many",
    method: (params) => {
      const values = {};
      for (let i = 0; i < len; i++) {
        values[foreignKeys[i]] = params[primaryKeys[i]];
      }
      return _queryDefaults(query.where(values), values);
    },
    virtualColumn: new HasManyVirtualColumn(
      defaultSchemaConfig,
      relationName,
      state
    ),
    joinQuery: joinQueryChainingHOF(
      reverseJoin,
      (joiningQuery, baseQuery) => joinHasRelation(baseQuery, joiningQuery, primaryKeys, foreignKeys, len)
    ),
    reverseJoin,
    modifyRelatedQuery(relationQuery) {
      return (query2) => {
        const baseQuery = query2.clone();
        baseQuery.q.select = fromQuerySelect;
        const q = relationQuery.q;
        q.kind = "from";
        q.values = { from: baseQuery };
      };
    }
  };
};
const getWhereForNestedUpdate = (t, data, params, primaryKeys, foreignKeys) => {
  return t.where({
    IN: {
      columns: foreignKeys,
      values: data.map((item) => primaryKeys.map((key) => item[key]))
    },
    OR: params ? toArray(params) : void 0
  });
};
const nestedInsert$1 = ({ query, primaryKeys, foreignKeys }) => {
  const len = primaryKeys.length;
  return async (_, data) => {
    const t = query.clone();
    const items = [];
    for (const item of data) {
      if (item[1].connect) {
        items.push(item);
      }
    }
    if (items.length) {
      for (let i = 0, len2 = items.length; i < len2; i++) {
        const [selfData, { connect }] = items[i];
        const obj = {};
        for (let i2 = 0; i2 < len2; i2++) {
          obj[foreignKeys[i2]] = selfData[primaryKeys[i2]];
        }
        items[i] = _queryUpdateOrThrow(
          t.orWhere(...connect),
          obj
        );
      }
      await Promise.all(items);
    }
    items.length = 0;
    for (const item of data) {
      if (item[1].connectOrCreate) {
        items.push(item);
      }
    }
    let connected;
    if (items.length) {
      const queries = [];
      for (let i = 0, len2 = items.length; i < len2; i++) {
        const [selfData, { connectOrCreate }] = items[i];
        for (const item of connectOrCreate) {
          const obj = {};
          for (let i2 = 0; i2 < len2; i2++) {
            obj[foreignKeys[i2]] = selfData[primaryKeys[i2]];
          }
          queries.push(
            _queryUpdate(t.where(item.where), obj)
          );
        }
      }
      connected = await Promise.all(queries);
    } else {
      connected = [];
    }
    let connectedI = 0;
    items.length = 0;
    for (const item of data) {
      if (item[1].connectOrCreate) {
        const length = item[1].connectOrCreate.length;
        connectedI += length;
        for (let i = length; i > 0; i--) {
          if (connected[connectedI - i] === 0) {
            items.push(item);
            break;
          }
        }
      } else if (item[1].create) {
        items.push(item);
      }
    }
    connectedI = 0;
    if (items.length) {
      const records = [];
      for (const [selfData, { create, connectOrCreate }] of items) {
        const obj = {};
        for (let i = 0; i < len; i++) {
          obj[foreignKeys[i]] = selfData[primaryKeys[i]];
        }
        if (create) {
          for (const item of create) {
            records.push(__spreadValues$4(__spreadValues$4({}, item), obj));
          }
        }
        if (connectOrCreate) {
          for (const item of connectOrCreate) {
            if (connected[connectedI++] === 0) {
              records.push(__spreadValues$4(__spreadValues$4({}, item.create), obj));
            }
          }
        }
      }
      await _queryCreateMany(t, records);
    }
  };
};
const nestedUpdate$1 = ({ query, primaryKeys, foreignKeys }) => {
  const len = primaryKeys.length;
  return async (_, data, params) => {
    var _a;
    const t = query.clone();
    if (params.create) {
      const obj = {};
      for (let i = 0; i < len; i++) {
        obj[foreignKeys[i]] = data[0][primaryKeys[i]];
      }
      await t.insertMany(
        params.create.map((create) => __spreadValues$4(__spreadValues$4({}, create), obj))
      );
      delete t.q[toSQLCacheKey];
    }
    if (params.disconnect || params.set) {
      const obj = {};
      for (const foreignKey of foreignKeys) {
        obj[foreignKey] = null;
      }
      await _queryUpdate(
        getWhereForNestedUpdate(
          t,
          data,
          params.disconnect,
          primaryKeys,
          foreignKeys
        ),
        obj
      );
      if (params.set && (Array.isArray(params.set) ? params.set.length : objectHasValues(params.set))) {
        delete t.q[toSQLCacheKey];
        const obj2 = {};
        for (let i = 0; i < len; i++) {
          obj2[foreignKeys[i]] = data[0][primaryKeys[i]];
        }
        await _queryUpdate(
          t.where(
            Array.isArray(params.set) ? {
              OR: params.set
            } : params.set
          ),
          obj2
        );
      }
    }
    if (params.delete || params.update) {
      delete t.q[toSQLCacheKey];
      const q = getWhereForNestedUpdate(
        t,
        data,
        params.delete || ((_a = params.update) == null ? void 0 : _a.where),
        primaryKeys,
        foreignKeys
      );
      if (params.delete) {
        await _queryDelete(q);
      } else if (params.update) {
        await _queryUpdate(q, params.update.data);
      }
    }
  };
};

var __defProp$3 = Object.defineProperty;
var __defProps$1 = Object.defineProperties;
var __getOwnPropDescs$1 = Object.getOwnPropertyDescriptors;
var __getOwnPropSymbols$3 = Object.getOwnPropertySymbols;
var __hasOwnProp$3 = Object.prototype.hasOwnProperty;
var __propIsEnum$3 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$3 = (obj, key, value) => key in obj ? __defProp$3(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$3 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$3.call(b, prop))
      __defNormalProp$3(a, prop, b[prop]);
  if (__getOwnPropSymbols$3)
    for (var prop of __getOwnPropSymbols$3(b)) {
      if (__propIsEnum$3.call(b, prop))
        __defNormalProp$3(a, prop, b[prop]);
    }
  return a;
};
var __spreadProps$1 = (a, b) => __defProps$1(a, __getOwnPropDescs$1(b));
class HasAndBelongsToManyVirtualColumn extends VirtualColumn {
  constructor(joinTable, schema, key, state) {
    super(schema);
    this.joinTable = joinTable;
    this.key = key;
    this.state = state;
    this.nestedInsert = nestedInsert(state);
    this.nestedUpdate = nestedUpdate(state);
  }
  create(q, ctx, item, rowIndex) {
    hasRelationHandleCreate(
      q,
      ctx,
      item,
      rowIndex,
      this.key,
      this.state.primaryKeys,
      this.nestedInsert
    );
  }
  update(q, _, set) {
    hasRelationHandleUpdate(
      q,
      set,
      this.key,
      this.state.primaryKeys,
      this.nestedUpdate
    );
  }
}
const removeColumnName = (column) => {
  if (!column.data.name)
    return column;
  const cloned = Object.create(column);
  cloned.data = __spreadProps$1(__spreadValues$3({}, column.data), { name: void 0 });
  return cloned;
};
const makeHasAndBelongsToManyMethod = (table, qb, relation, relationName, query) => {
  let primaryKeys;
  let foreignKeys;
  let joinTable;
  let throughForeignKeys;
  let throughPrimaryKeys;
  const { options } = relation;
  if ("columns" in options) {
    primaryKeys = options.columns;
    foreignKeys = options.references;
    joinTable = options.through.table;
    throughForeignKeys = options.through.columns;
    throughPrimaryKeys = options.through.references;
  } else {
    primaryKeys = [options.primaryKey];
    foreignKeys = [options.foreignKey];
    joinTable = options.joinTable;
    throughForeignKeys = [options.associationForeignKey];
    throughPrimaryKeys = [options.associationPrimaryKey];
  }
  const { snakeCase } = table.internal;
  const foreignKeysFull = foreignKeys.map((key, i) => {
    if (snakeCase)
      key = foreignKeys[i] = toSnakeCase(key);
    return `${joinTable}.${key}`;
  });
  const throughForeignKeysFull = throughForeignKeys.map((key, i) => {
    if (snakeCase)
      key = throughForeignKeys[i] = toSnakeCase(key);
    return `${joinTable}.${key}`;
  });
  const foreignTable = getQueryAs(query);
  const throughPrimaryKeysFull = throughPrimaryKeys.map(
    (key) => `${foreignTable}.${key}`
  );
  const len = primaryKeys.length;
  const throughLen = throughPrimaryKeys.length;
  const baseQuery = Object.create(qb.baseQuery);
  baseQuery.baseQuery = baseQuery;
  baseQuery.table = joinTable;
  const shape = {};
  for (let i = 0; i < len; i++) {
    shape[foreignKeys[i]] = removeColumnName(
      table.shape[primaryKeys[i]]
    );
  }
  for (let i = 0; i < throughLen; i++) {
    shape[throughForeignKeys[i]] = removeColumnName(
      query.shape[throughPrimaryKeys[i]]
    );
  }
  baseQuery.shape = shape;
  baseQuery.q = __spreadProps$1(__spreadValues$3({}, baseQuery.q), {
    shape: baseQuery.shape
  });
  const subQuery = Object.create(baseQuery);
  const state = {
    relatedTableQuery: query,
    joinTableQuery: subQuery,
    primaryKeys,
    foreignKeys,
    throughForeignKeys,
    throughPrimaryKeys,
    foreignKeysFull,
    throughForeignKeysFull,
    throughPrimaryKeysFull
  };
  const joinQuery = (joiningQuery, tableAs, foreignAs, joinedShapes) => {
    const cloned = joiningQuery.clone();
    cloned.q.joinedShapes = joinedShapes;
    return _queryWhereExists(cloned, subQuery, [
      (q) => {
        for (let i = 0; i < throughLen; i++) {
          _queryJoinOn(q, [
            throughForeignKeysFull[i],
            `${foreignAs}.${throughPrimaryKeys[i]}`
          ]);
        }
        for (let i = 0; i < len; i++) {
          _queryJoinOn(q, [foreignKeysFull[i], `${tableAs}.${primaryKeys[i]}`]);
        }
        return q;
      }
    ]);
  };
  const obj = {};
  for (let i = 0; i < len; i++) {
    obj[foreignKeys[i]] = primaryKeys[i];
  }
  const selectPrimaryKeysAsForeignKeys = [{ selectAs: obj }];
  const reverseJoin = (baseQuery2, joiningQuery) => {
    const foreignAs = getQueryAs(joiningQuery);
    return joinQuery(baseQuery2, getQueryAs(baseQuery2), foreignAs, __spreadProps$1(__spreadValues$3({}, baseQuery2.q.joinedShapes), {
      [foreignAs]: joiningQuery.q.shape
    }));
  };
  return {
    returns: "many",
    method(params) {
      return query.whereExists(subQuery, (q) => {
        q = q.clone();
        const where = {};
        for (let i = 0; i < len; i++) {
          where[foreignKeysFull[i]] = params[primaryKeys[i]];
        }
        for (let i = 0; i < throughLen; i++) {
          _queryJoinOn(q, [
            throughForeignKeysFull[i],
            throughPrimaryKeysFull[i]
          ]);
        }
        return _queryWhere(q, [where]);
      });
    },
    virtualColumn: new HasAndBelongsToManyVirtualColumn(
      subQuery,
      defaultSchemaConfig,
      relationName,
      state
    ),
    joinQuery: joinQueryChainingHOF(
      reverseJoin,
      (joiningQuery, baseQuery2) => joinQuery(joiningQuery, getQueryAs(baseQuery2), getQueryAs(joiningQuery), __spreadProps$1(__spreadValues$3({}, joiningQuery.q.joinedShapes), {
        [baseQuery2.q.as || baseQuery2.table]: baseQuery2.q.shape
      }))
    ),
    reverseJoin,
    modifyRelatedQuery(relationQuery) {
      const ref = {};
      _queryHookAfterCreate(relationQuery, [], async (result) => {
        if (result.length > 1) {
          throw new OrchidOrmInternalError(
            relationQuery,
            "Creating multiple `hasAndBelongsToMany` records is not yet supported"
          );
        }
        const baseQuery2 = ref.q.clone();
        baseQuery2.q.select = selectPrimaryKeysAsForeignKeys;
        const data = {};
        for (let i = 0; i < throughLen; i++) {
          data[throughForeignKeys[i]] = result[0][throughPrimaryKeys[i]];
        }
        const createdCount = await _queryCreateFrom(
          subQuery.count(),
          baseQuery2,
          data
        );
        if (createdCount === 0) {
          throw new NotFoundError(baseQuery2);
        }
      });
      return (q) => {
        ref.q = q;
      };
    }
  };
};
const queryJoinTable = (state, data, conditions) => {
  const t = state.joinTableQuery.where({
    IN: {
      columns: state.foreignKeys,
      values: data.map((item) => state.primaryKeys.map((key) => item[key]))
    }
  });
  if (conditions) {
    _queryWhere(t, [
      {
        IN: {
          columns: state.throughForeignKeys,
          values: _querySelect(
            state.relatedTableQuery.where(conditionsToWhereArg(conditions)),
            state.throughPrimaryKeys
          )
        }
      }
    ]);
  }
  return t;
};
const conditionsToWhereArg = (conditions) => Array.isArray(conditions) ? { OR: conditions } : conditions;
const insertToJoinTable = (state, joinTableTransaction, data, idsRows) => {
  const len = state.primaryKeys.length;
  const throughLen = state.throughPrimaryKeys.length;
  const records = [];
  for (const item of data) {
    const obj = {};
    for (let i = 0; i < len; i++) {
      obj[state.foreignKeys[i]] = item[state.primaryKeys[i]];
    }
    for (const ids of idsRows) {
      const record = __spreadValues$3({}, obj);
      for (let i = 0; i < throughLen; i++) {
        record[state.throughForeignKeys[i]] = ids[i];
      }
      records.push(record);
    }
  }
  return joinTableTransaction.insertMany(records);
};
const nestedInsert = ({
  relatedTableQuery,
  joinTableQuery,
  primaryKeys,
  foreignKeys,
  throughPrimaryKeys,
  throughForeignKeys
}) => {
  const len = primaryKeys.length;
  const throughLen = primaryKeys.length;
  return async (_, data) => {
    const t = relatedTableQuery.clone();
    const items = [];
    for (const item of data) {
      if (item[1].connect) {
        items.push(item);
      }
    }
    let connected;
    if (items.length) {
      const queries = [];
      for (const [, { connect }] of items) {
        for (const item of connect) {
          queries.push(
            _queryFindBy(t.select(...throughPrimaryKeys), [
              item
            ])
          );
        }
      }
      connected = await Promise.all(queries);
    } else {
      connected = [];
    }
    items.length = 0;
    for (const item of data) {
      if (item[1].connectOrCreate) {
        items.push(item);
      }
    }
    let connectOrCreated;
    if (items.length) {
      const queries = [];
      for (const [, { connectOrCreate }] of items) {
        for (const item of connectOrCreate) {
          queries.push(
            _queryFindByOptional(t.select(...throughPrimaryKeys), [
              item.where
            ])
          );
        }
      }
      connectOrCreated = await Promise.all(queries);
    } else {
      connectOrCreated = [];
    }
    let connectOrCreateI = 0;
    items.length = 0;
    for (const item of data) {
      if (item[1].connectOrCreate) {
        const length = item[1].connectOrCreate.length;
        connectOrCreateI += length;
        for (let i = length; i > 0; i--) {
          if (!connectOrCreated[connectOrCreateI - i]) {
            items.push(item);
            break;
          }
        }
      } else if (item[1].create) {
        items.push(item);
      }
    }
    connectOrCreateI = 0;
    let created;
    if (items.length) {
      const records2 = [];
      for (const [, { create, connectOrCreate }] of items) {
        if (create) {
          records2.push(...create);
        }
        if (connectOrCreate) {
          for (const item of connectOrCreate) {
            if (!connectOrCreated[connectOrCreateI++]) {
              records2.push(item.create);
            }
          }
        }
      }
      created = await _queryCreateMany(
        t.select(...throughPrimaryKeys),
        records2
      );
    } else {
      created = [];
    }
    const allKeys = data;
    let createI = 0;
    let connectI = 0;
    connectOrCreateI = 0;
    for (let index = 0, len2 = data.length; index < len2; index++) {
      const item = data[index][1];
      if (item.create || item.connectOrCreate) {
        if (item.create) {
          const len3 = item.create.length;
          allKeys[index][1] = created.slice(createI, createI + len3);
          createI += len3;
        }
        if (item.connectOrCreate) {
          const arr = [];
          allKeys[index][1] = arr;
          const len3 = item.connectOrCreate.length;
          for (let i = 0; i < len3; i++) {
            const item2 = connectOrCreated[connectOrCreateI++];
            if (item2) {
              arr.push(item2);
            } else {
              arr.push(created[createI++]);
            }
          }
        }
      }
      if (item.connect) {
        const len3 = item.connect.length;
        allKeys[index][1] = connected.slice(connectI, connectI + len3);
        connectI += len3;
      }
    }
    const records = [];
    for (const [selfData, relationKeys] of allKeys) {
      const obj = {};
      for (let i = 0; i < len; i++) {
        obj[foreignKeys[i]] = selfData[primaryKeys[i]];
      }
      for (const relationData of relationKeys) {
        const record = __spreadValues$3({}, obj);
        for (let i = 0; i < throughLen; i++) {
          record[throughForeignKeys[i]] = relationData[throughPrimaryKeys[i]];
        }
        records.push(record);
      }
    }
    await joinTableQuery.insertMany(records);
  };
};
const nestedUpdate = (state) => {
  const len = state.primaryKeys.length;
  const throughLen = state.throughPrimaryKeys.length;
  return async (_, data, params) => {
    if (params.create) {
      const idsRows = await _queryCreateMany(
        _queryRows(state.relatedTableQuery.select(...state.throughPrimaryKeys)),
        params.create
      );
      const records = [];
      for (const item of data) {
        const obj = {};
        for (let i = 0; i < len; i++) {
          obj[state.foreignKeys[i]] = item[state.primaryKeys[i]];
        }
        for (const ids of idsRows) {
          const record = __spreadValues$3({}, obj);
          for (let i = 0; i < throughLen; i++) {
            record[state.throughForeignKeys[i]] = ids[i];
          }
          records.push(record);
        }
      }
      await state.joinTableQuery.createMany(records);
    }
    if (params.update) {
      await _queryUpdate(
        _queryWhere(
          state.relatedTableQuery.whereExists(state.joinTableQuery, (q) => {
            for (let i = 0; i < throughLen; i++) {
              _queryJoinOn(q, [
                state.throughForeignKeysFull[i],
                state.throughPrimaryKeysFull[i]
              ]);
            }
            return _queryWhere(q, [
              {
                IN: {
                  columns: state.foreignKeysFull,
                  values: data.map(
                    (item) => state.primaryKeys.map((key) => item[key])
                  )
                }
              }
            ]);
          }),
          [conditionsToWhereArg(params.update.where)]
        ),
        params.update.data
      );
    }
    if (params.disconnect) {
      await _queryDelete(
        queryJoinTable(state, data, params.disconnect)
      );
    }
    if (params.delete) {
      const j = queryJoinTable(state, data, params.delete);
      const idsRows = await _queryDelete(
        _queryRows(_querySelect(j, state.throughForeignKeys))
      );
      await _queryDelete(
        state.relatedTableQuery.where({
          IN: {
            columns: state.throughPrimaryKeys,
            values: idsRows
          }
        })
      );
    }
    if (params.set) {
      const j = queryJoinTable(state, data);
      await _queryDelete(j);
      delete j.q[toSQLCacheKey];
      if (Array.isArray(params.set) ? params.set.length : objectHasValues(params.set)) {
        const idsRows = await _queryRows(
          _querySelect(
            state.relatedTableQuery.where(
              conditionsToWhereArg(params.set)
            ),
            state.throughPrimaryKeys
          )
        );
        await insertToJoinTable(state, j, data, idsRows);
      }
    }
  };
};

var __defProp$2 = Object.defineProperty;
var __getOwnPropSymbols$2 = Object.getOwnPropertySymbols;
var __hasOwnProp$2 = Object.prototype.hasOwnProperty;
var __propIsEnum$2 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$2 = (obj, key, value) => key in obj ? __defProp$2(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$2 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$2.call(b, prop))
      __defNormalProp$2(a, prop, b[prop]);
  if (__getOwnPropSymbols$2)
    for (var prop of __getOwnPropSymbols$2(b)) {
      if (__propIsEnum$2.call(b, prop))
        __defNormalProp$2(a, prop, b[prop]);
    }
  return a;
};
const applyRelations = (qb, tables, result) => {
  var _a;
  const tableEntries = Object.entries(tables);
  const delayedRelations = /* @__PURE__ */ new Map();
  for (const name in tables) {
    const table = tables[name];
    if (!("relations" in table) || typeof table.relations !== "object")
      continue;
    const dbTable = result[name];
    for (const relationName in table.relations) {
      const relation = table.relations[relationName];
      const otherTableClass = relation.fn();
      const otherTable = tableEntries.find(
        (pair) => pair[1] instanceof otherTableClass
      );
      if (!otherTable) {
        throw new Error(
          `Cannot find table class for class ${otherTableClass.name}`
        );
      }
      const otherTableName = otherTable[0];
      const otherDbTable = result[otherTableName];
      if (!otherDbTable)
        throw new Error(`Cannot find table class by name ${otherTableName}`);
      const data = {
        relationName,
        relation,
        dbTable,
        otherDbTable
      };
      const options = relation.options;
      if (typeof options.through === "string" && typeof options.source === "string") {
        const throughRelation = getThroughRelation(dbTable, options.through);
        if (!throughRelation) {
          delayRelation(delayedRelations, dbTable, options.through, data);
          continue;
        }
        const sourceRelation = getSourceRelation(
          throughRelation,
          options.source
        );
        if (!sourceRelation) {
          delayRelation(
            delayedRelations,
            throughRelation.table,
            options.source,
            data
          );
          continue;
        }
      }
      applyRelation(qb, data, delayedRelations);
    }
  }
  if (delayedRelations.size) {
    const { value } = delayedRelations.values().next();
    for (const key in value) {
      for (const item of value[key]) {
        const { relation } = item;
        if (item.dbTable.relations[item.relationName])
          continue;
        const as = item.dbTable.definedAs;
        let message = `Cannot define a \`${item.relationName}\` relation on \`${as}\``;
        const table = result[as];
        const { through, source } = relation.options;
        const throughRel = (_a = table.relations[through]) == null ? void 0 : _a.relationConfig;
        if (through && !throughRel) {
          message += `: cannot find \`${through}\` relation required by the \`through\` option`;
        } else if (source && throughRel && !throughRel.table.relations[source]) {
          message += `: cannot find \`${source}\` relation in \`${throughRel.table.definedAs}\` required by the \`source\` option`;
        }
        throw new Error(message);
      }
    }
  }
};
const delayRelation = (delayedRelations, table, relationName, data) => {
  let tableRelations = delayedRelations.get(table);
  if (!tableRelations) {
    tableRelations = {};
    delayedRelations.set(table, tableRelations);
  }
  if (tableRelations[relationName]) {
    tableRelations[relationName].push(data);
  } else {
    tableRelations[relationName] = [data];
  }
};
const applyRelation = (qb, { relationName, relation, dbTable, otherDbTable }, delayedRelations) => {
  var _a;
  const baseQuery = Object.create(otherDbTable);
  baseQuery.baseQuery = baseQuery;
  const query = (relation.options.scope ? relation.options.scope(baseQuery) : baseQuery).as(relationName);
  const definedAs = query.definedAs;
  if (!definedAs) {
    throw new Error(
      `Table class for table ${query.table} is not attached to db instance`
    );
  }
  const { type } = relation;
  let data;
  if (type === "belongsTo") {
    data = makeBelongsToMethod(relation, relationName, query);
  } else if (type === "hasOne") {
    data = makeHasOneMethod(dbTable, relation, relationName, query);
  } else if (type === "hasMany") {
    data = makeHasManyMethod(dbTable, relation, relationName, query);
  } else if (type === "hasAndBelongsToMany") {
    data = makeHasAndBelongsToManyMethod(
      dbTable,
      qb,
      relation,
      relationName,
      query
    );
  } else {
    throw new Error(`Unknown relation type ${type}`);
  }
  if (data.returns === "one") {
    if (relation.options.required) {
      _queryTake(query);
    } else {
      _queryTakeOptional(query);
    }
    query.q.returnsOne = true;
  }
  if (data.virtualColumn) {
    dbTable.shape[relationName] = dbTable.q.shape[relationName] = data.virtualColumn;
  }
  makeRelationQuery(dbTable, relationName, data, query);
  baseQuery.joinQuery = data.joinQuery;
  const { join: originalJoin } = baseQuery;
  baseQuery.join = function(...args) {
    if (args.length) {
      return originalJoin.apply(this, args);
    } else {
      const q = this.clone();
      q.q.innerJoinLateral = true;
      return q;
    }
  };
  baseQuery.relationConfig = {
    table: otherDbTable,
    query,
    joinQuery: data.joinQuery
  };
  dbTable.relations[relationName] = query;
  const tableRelations = delayedRelations.get(dbTable);
  if (!tableRelations)
    return;
  (_a = tableRelations[relationName]) == null ? void 0 : _a.forEach((data2) => {
    applyRelation(qb, data2, delayedRelations);
  });
};
const makeRelationQuery = (table, relationName, data, q) => {
  Object.defineProperty(table, relationName, {
    configurable: true,
    get() {
      var _a;
      const toTable = q.clone();
      let query;
      if (this.q.isSubQuery) {
        query = toTable;
        query.q.isSubQuery = true;
      } else {
        query = _queryWhere(_queryAll(toTable), [
          {
            EXISTS: { q: data.reverseJoin(this, toTable) }
          }
        ]);
      }
      if (this.q.relChain) {
        query.q.relChain = [...this.q.relChain, this];
        query.q.returnType = "all";
      } else {
        query.q.relChain = [this];
      }
      query.q.joinedShapes = __spreadValues$2({
        [getQueryAs(this)]: this.q.shape
      }, this.q.joinedShapes);
      const setQuery = (_a = data.modifyRelatedQuery) == null ? void 0 : _a.call(data, query);
      setQuery == null ? void 0 : setQuery(this);
      return new Proxy(data.method, {
        get(_, prop) {
          return query[prop];
        }
      });
    },
    set(value) {
      Object.defineProperty(this, relationName, {
        value
      });
    }
  });
};

function transaction(fnOrOptions, fn) {
  return this.$queryBuilder.transaction(
    fnOrOptions,
    fn
  );
}

var __defProp$1 = Object.defineProperty;
var __defProps = Object.defineProperties;
var __getOwnPropDescs = Object.getOwnPropertyDescriptors;
var __getOwnPropSymbols$1 = Object.getOwnPropertySymbols;
var __hasOwnProp$1 = Object.prototype.hasOwnProperty;
var __propIsEnum$1 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$1 = (obj, key, value) => key in obj ? __defProp$1(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$1 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$1.call(b, prop))
      __defNormalProp$1(a, prop, b[prop]);
  if (__getOwnPropSymbols$1)
    for (var prop of __getOwnPropSymbols$1(b)) {
      if (__propIsEnum$1.call(b, prop))
        __defNormalProp$1(a, prop, b[prop]);
    }
  return a;
};
var __spreadProps = (a, b) => __defProps(a, __getOwnPropDescs(b));
var __objRest = (source, exclude) => {
  var target = {};
  for (var prop in source)
    if (__hasOwnProp$1.call(source, prop) && exclude.indexOf(prop) < 0)
      target[prop] = source[prop];
  if (source != null && __getOwnPropSymbols$1)
    for (var prop of __getOwnPropSymbols$1(source)) {
      if (exclude.indexOf(prop) < 0 && __propIsEnum$1.call(source, prop))
        target[prop] = source[prop];
    }
  return target;
};
const orchidORM = (_a, tables) => {
  var _b = _a, {
    log,
    logger,
    autoPreparedStatements,
    noPrimaryKey = "error"
  } = _b, options = __objRest(_b, [
    "log",
    "logger",
    "autoPreparedStatements",
    "noPrimaryKey"
  ]);
  var _a2;
  const commonOptions = {
    log,
    logger,
    autoPreparedStatements,
    noPrimaryKey
  };
  let adapter;
  let transactionStorage;
  let qb;
  if ("db" in options) {
    adapter = options.db.q.adapter;
    transactionStorage = options.db.internal.transactionStorage;
    qb = options.db.queryBuilder;
  } else {
    adapter = "adapter" in options ? options.adapter : new Adapter(options);
    transactionStorage = new AsyncLocalStorage();
    qb = _initQueryBuilder(
      adapter,
      makeColumnTypes(defaultSchemaConfig),
      transactionStorage,
      commonOptions,
      options
    );
  }
  const result = {
    $transaction: transaction,
    $adapter: adapter,
    $queryBuilder: qb,
    $query: (...args) => qb.query(...args),
    $queryArrays: (...args) => qb.queryArrays(...args),
    $with: qb.with.bind(qb),
    $withRecursive: qb.withRecursive.bind(qb),
    $withSql: qb.withSql.bind(qb),
    $from: qb.from.bind(qb),
    $close: adapter.close.bind(adapter)
  };
  const tableInstances = {};
  for (const key in tables) {
    if (key[0] === "$") {
      throw new Error(`Table class name must not start with $`);
    }
    const table = tables[key].instance();
    tableInstances[key] = table;
    const options2 = __spreadProps(__spreadValues$1({}, commonOptions), {
      schema: table.schema,
      language: table.language,
      scopes: table.scopes,
      softDelete: table.softDelete,
      snakeCase: table.snakeCase,
      comment: table.comment,
      noPrimaryKey: table.noPrimaryKey ? "ignore" : void 0
    });
    const dbTable = new Db(
      adapter,
      // eslint-disable-next-line @typescript-eslint/no-explicit-any
      qb,
      table.table,
      table.columns.shape,
      table.types,
      transactionStorage,
      options2,
      (_a2 = table.constructor.prototype.tableData) != null ? _a2 : emptyObject
    );
    dbTable.definedAs = key;
    dbTable.db = result;
    dbTable.filePath = table.filePath;
    dbTable.name = table.constructor.name;
    if (table.computed) {
      addComputedColumns(
        dbTable,
        table.computed
      );
    }
    result[key] = dbTable;
  }
  applyRelations(qb, tableInstances, result);
  for (const key in tables) {
    const table = tableInstances[key];
    if (table.init) {
      table.init(result);
      Object.assign(result[key].baseQuery.q, table.q);
    }
  }
  return result;
};

var __defProp = Object.defineProperty;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
const createRepo = (table, methods) => {
  const queryMethods = __spreadValues(__spreadValues(__spreadValues(__spreadValues({}, methods.queryMethods), methods.queryOneMethods), methods.queryWithWhereMethods), methods.queryOneWithWhereMethods);
  const plainMethods = methods.methods;
  const repo = (q2) => {
    const proto = Object.create(q2.baseQuery);
    proto.baseQuery = proto;
    const result = Object.create(proto);
    result.q = getClonedQueryData(q2.q);
    if (plainMethods) {
      Object.assign(proto.baseQuery, plainMethods);
    }
    for (const key in queryMethods) {
      const method = queryMethods[key];
      proto.baseQuery[key] = function(...args) {
        return method(this, ...args);
      };
    }
    return result;
  };
  const q = repo(table);
  return new Proxy(repo, {
    get(_, key) {
      return q[key];
    }
  });
};

export { createBaseTable, createRepo, orchidORM };
//# sourceMappingURL=index.mjs.map
